# Script for creating json with annotations in the COCO format out of the Cell Tracking Challenge format of annotations
# Author: Kristyna Janku

import json
import cv2
import glob
import numpy as np
from os import listdir, path


def get_image_info(videos_folder, video, removed_frames):
    images = []

    img_id = 0
    video_path = path.join(videos_folder, video)

    # Filter out frames that were removed due to incomplete SEG annotations
    frame_paths = sorted(frame for frame in glob.glob(path.join(video_path, '*.png'))
                         if ''.join(filter(str.isdigit, path.basename(frame))) not in removed_frames)

    for filepath in frame_paths:
        img = cv2.imread(filepath)
        height, width, _ = img.shape
        images.append(
            {
                'file_name': path.basename(filepath),
                'height': height,
                'id': img_id,
                'width': width
            }
        )
        img_id += 1

    return images


def get_annotations(videos_folder, video):
    annotations = []

    gt_annotation_path = path.join(videos_folder, f'{video}_GT')
    gt_segmentation_folder_path = path.join(gt_annotation_path, 'SEG')
    gt_tracking_file_path = path.join(gt_annotation_path, 'TRA', 'man_track.txt')

    st_segmentation_folder_path = path.join(videos_folder, f'{video}_ST', 'SEG')

    # Load numbers of all frames from filenames (only PNG files)
    frames = sorted([''.join(filter(str.isdigit, frame)) for frame
                     in listdir(path.join(videos_folder, video)) if frame.endswith('.png')])

    removed_frames = []
    id_set = []
    id_frames = {}

    # Load tracking file and get information about objects and frames they appear in
    with open(gt_tracking_file_path, 'r') as track_file:
        for line in track_file:
            parts = line.split(' ')
            obj_id = parts[0]
            id_set.append(int(obj_id))
            id_frames[int(obj_id)] = list(range(int(parts[1]), (int(parts[2]) + 1)))

    for frame in frames:
        frame_annotations = []
        # Load segmented image, gold truth if available, silver truth otherwise
        seg = glob.glob(path.join(gt_segmentation_folder_path, f'*{frame}*'))
        if len(seg) == 0:
            seg = glob.glob(path.join(st_segmentation_folder_path, f'*{frame}*'))
        seg_image = cv2.imread(seg[0], cv2.IMREAD_ANYDEPTH)

        for obj_id in range(1, seg_image.max() + 1):
            if obj_id in id_set and int(frame) in id_frames[obj_id]:
                # Mask only pixels with current object, multiply by 1 to change True/False to 1/0
                mask = (seg_image == obj_id) * 1
                # Convert to 8-bit image to enable further transformations
                mask = cv2.convertScaleAbs(mask)
                # Checking if tracked object really is in segmented image
                if mask.max() == 0:
                    print(f'Video: {video} Frame: {frame} Missing object id in seg: {str(obj_id)}')
                    frame_annotations = []
                    removed_frames.append(frame)
                    break
                else:
                    # Get bounding box of object from contour
                    contours = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[0]
                    areas = [cv2.contourArea(c) for c in contours]
                    max_index = np.argmax(areas)
                    max_contour = contours[max_index]
                    x, y, width, height = cv2.boundingRect(max_contour)

                    # Add padding
                    x = max(x - 5, 0)
                    y = max(y - 5, 0)
                    width = width + 10
                    height = height + 10

                    area = width * height

                    # Ignore annotations with only a few pixels
                    if area > 5:
                        frame_annotations.append(
                            {
                                'area': area,
                                'bbox': [x, y, width, height],
                                'category_id': 0,
                                'id': obj_id - 1,
                                'ignore': 0,
                                'image_id': int(frame)
                            }
                        )
        annotations += frame_annotations
    return annotations, removed_frames


def create_json(dataset_name, videos_folder, videos):
    for video in videos:
        # Add category info
        categories = [
                {
                    'id': 0,
                    'name': 'cell',
                    'supercategory': 'cell'
                }
            ]

        annotations, removed_frames = get_annotations(videos_folder, video)

        coco_dict = {
            'categories': categories,
            'annotations': annotations,
            'images': get_image_info(videos_folder, video, removed_frames),
        }

        with open(f'{dataset_name}_video_{video}.json', 'w') as ann_json_file:
            json.dump(coco_dict, ann_json_file, indent=4, sort_keys=True)


if __name__ == '__main__':
    def_dataset_name = 'BF-C2DL-MuSC'
    def_videos_folder = '/home/kika/data/BF-C2DL-MuSC/train/data'
    def_videos = ['01']
    create_json(def_dataset_name, def_videos_folder, def_videos)

